/* Simple utility to generate a source file containing the raw VDSO image that
 * we compile into the kernel
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#define die(fmt, args...) {\
    fprintf(stderr, fmt, ##args);\
    exit(EXIT_FAILURE);\
}

static void
write_header(FILE * fp)
{
    fprintf(fp, "/* This file is automatically generated by vdsogen\n");
    fprintf(fp, " * - DO NOT MODIFY -\n");
    fprintf(fp, " */\n");

    fprintf(fp, "unsigned char vdso_raw[8192] __attribute__((aligned (4096))) = {");
}

static void
write_footer(FILE * fp)
{
    fprintf(fp, "\n");
    fprintf(fp, "};\n");
}

int main(int argc, char ** argv)
{
    void * map;
    int fd;
    FILE * fp_out;
    unsigned long total;

    struct stat st;

    if (argc != 3)
        die("usage: %s <vdso.so> <output.c>\n", *argv);

    fd = open(argv[1], O_RDONLY);
    if (fd == -1)
        die("failed to open %s: %s\n", argv[1], strerror(errno));

    if (fstat(fd, &st) == -1)
        die("failed to stat %s: %s\n", argv[1], strerror(errno));

    map = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
    if (map == MAP_FAILED)
        die("failed to mmap %s: %s\n", argv[1], strerror(errno));

    fp_out = fopen(argv[2], "w");
    if (fp_out == NULL)
        die("failed to open %s: %s\n", argv[2], strerror(errno));

    write_header(fp_out);

    for (total = 0; ; total++) {
        unsigned char byte;
        ssize_t ret;

        ret = read(fd, &byte, sizeof(unsigned char));
        if (ret == -1)
            die("failed to read %s: %s\n", argv[1], strerror(errno));

        if (ret == 0)
            break;

        if (ret != sizeof(unsigned char))
            die("read %ld bytes instead of %ld\n", ret, sizeof(unsigned char));

        fprintf(fp_out, "%s0x%02X,%s", 
            (total % 8 == 0) ? "\n    " : "",
            byte,
            (total % 8 != 7) ? " " : ""
        );
    }

    write_footer(fp_out);
    fclose(fp_out);
    munmap(map, st.st_size);
    close(fd);

    exit(EXIT_SUCCESS);
}
